import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import processing.opengl.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class Artikulation_Player extends PApplet {

//artikulation player
//kevin schlei
//updated: 10.8.2018



ArrayList buffers;

float scrollRate;

//!
float tilterCounter;

PImage[] artikulationCropped;
float[] artikulationImageScalers;
int artikulationScoreCounter;
PImage textureImage;

PFont titleFont;
boolean isPlaying;
boolean playScore;
boolean playGenerative;

//globals:
int bgColor;
boolean showRects, showRandBufferBackground;
int minimumBufferWidth;
int[] allColors, dotColors, combColors;
int[] dotColorRandom, combColorRandom;
int dotColorCycler, combColorCycler, lastDotColorIndex, lastCombColorIndex;
int buffersGenerated, maxGeneratedBuffers;
float tiltTint;

public Shape getMainShape() {
  Shape s;
  float r = random(1);
  if (r >= 0.0f && r < 0.15f) {
    s = new DotCurvedTail();
    s.setSizes(40, 35, 90, 60);
  } else if (r >= 0.15f && r < 0.3f) {
    s = new DotStraightTail();
    s.setSizes(40, 20, 190, 32);
  } else if (r >= 0.3f && r < 0.4f) {
    s = new BlocksTailGroup();
    s.setSizes(200, 100, 500, PApplet.parseInt(height * 0.75f));
  } else if (r >= 0.4f && r < 0.5f) {
    s = new Comb();
    s.setSizes(20, 40, 40, 50);
  } else {
    s = new DotsConnected();
    s.setSizes(35, 60, 110, 110);
  }
  return s;
}

public Shape getLowerShape() {
  Shape s;
  s = new Wedges();
  s.setSizes(18, 40, 55, 70);
  return s;
}

public Shape getUpperShape() {
  Shape s;
  float r = random(1);
  if (r >= 0.0f && r <= 0.4f) {
    s = new Blocks();
    s.setSizes(20, 35, 60, 90);
  } else if (r >= 0.4f && r <= 0.5f) {
    s = new Block();
    s.setSizes(19, 19, 21, 21);
  } else if (r >= 0.5f && r <= 0.6f) {
    s = new Dot();
    s.setSizes(15, 15, 35, 35);
  } else {
    s = new Comb();
    s.setSizes(15, 25, 45, 70);
  }
  return s;
}

public int getDotColor() {
  if (dotColorCycler == 0) {
    for (int i=0; i<dotColors.length; i++) {
      if (true) lastDotColorIndex = PApplet.parseInt(random(dotColors.length));
      dotColorRandom[i] = lastDotColorIndex;
    }
  }
  int c = dotColors[dotColorRandom[dotColorCycler]];
  //  dotColorCycler = (dotColorCycler + 1) % dotColorRandom.length;

  return c;
}

public int getCombColor() {
  if (combColorCycler == 0) {
    for (int i=0; i<combColors.length; i++) {
      if (true) lastCombColorIndex = PApplet.parseInt(random(combColors.length));
      combColorRandom[i] = lastCombColorIndex;
    }
  }
  int c = combColors[combColorRandom[combColorCycler]];
  //  combColorCycler = (combColorCycler + 1) % combColorRandom.length;

  return c;
}

public void setup() {
  
  background(255);
  noCursor();
  frame.setBackground(new java.awt.Color(0, 0, 0));

  ///////////////////
  //DEFAULTS
  ///////////////////

  bgColor = color(255, 255, 255);
  showRects = false;
  showRandBufferBackground = false;
  scrollRate = 1.4f; //in pixels
  minimumBufferWidth = 100;
  maxGeneratedBuffers = 50;
  ///////////////////
  ///////////////////

  textureImage = loadImage("texture.png");

  dotColorRandom = new int[20];
  combColorRandom = new int[20];
  lastDotColorIndex = 0;
  lastCombColorIndex = 0;
  buffersGenerated = 0;
  tiltTint = 128.0f;

  allColors = new int[10];
  allColors[0] = color(0xff794a39); //dark brown
  allColors[1] = color(0xffdb0000); //red
  allColors[2] = color(0xffd45f00); //orange
  allColors[3] = color(0xffb99170); //light brown
  allColors[4] = color(0xffead442); //yellow
  allColors[5] = color(0xff94ad53); //light green
  allColors[6] = color(0xff5b8257); //dark green
  allColors[7] = color(0xff5c6396); //blue (origin 97,96,148
  allColors[8] = color(0xffba6891); //pink
  allColors[9] = color(0, 0, 0); //black

  dotColors = new int[6];
  dotColors[0] = allColors[1];
  dotColors[1] = allColors[2];
  dotColors[2] = allColors[4];
  dotColors[3] = allColors[7];
  dotColors[4] = allColors[8];
  dotColors[5] = allColors[9];

  combColors = new int[5];
  combColors[0] = allColors[0];
  combColors[1] = allColors[3];
  combColors[2] = allColors[5];
  combColors[3] = allColors[6];
  combColors[4] = allColors[9];

  //  frameRate(50);
  buffers = new ArrayList();

  artikulationCropped = new PImage[21];
  artikulationImageScalers = new float[21];
  for (int i=0; i<21; i++) {
    artikulationCropped[i] = loadImage("art-"+i+".png");
  }
  artikulationScoreCounter = 0;

  titleFont = createFont("Menlo", 32);
  textFont(titleFont, 32);

  noLoop(); //keypressed will start playback with options

  isPlaying = false;
  playScore = false;
  playGenerative = false;

  tilterCounter = 0;
}

public void draw() {
    background(255);
    println(frameRate);
    
  if (buffersGenerated > maxGeneratedBuffers) {
    //    float tilt = max(1, ((float(mouseY) / 4) / height) * 255.0);
    tiltTint -= 0.03f;
    scrollRate += 0.001f;
    tiltTint = max(1, tiltTint);
    //    blend(0,0,ADD);
    tint(255, tiltTint);
  }

  //if a buffer area is in range, display it:
  for (int i=0; i<buffers.size(); i++) {
    BufferArea buf = (BufferArea)buffers.get(i);
    if (buf.scrollPosition < width) {
      image(buf.displayImage, buf.scrollPosition, 0, buf.bufferWidth, height);
    }
    buf.scrollPosition -= scrollRate;
  }

  if (buffers.size() > 0) {
    //if it's scrolled out of sight, remove it:
    BufferArea firstBuf = (BufferArea)buffers.get(0);
    if (firstBuf.scrollPosition + firstBuf.bufferWidth < 0) {
      buffers.remove(0);
      //      println("remove buffer new list length is: "+buffers.size());
    }

    //check again that we're not dealing w/ empty array:
    if (buffers.size() > 0) {
      //if we need a new buffer to fill the screen, add it:
      firstBuf = (BufferArea)buffers.get(0);
      float totalWidth = firstBuf.scrollPosition;
      for (int i=0; i<buffers.size(); i++) {
        BufferArea buf = (BufferArea)buffers.get(i);
        totalWidth += buf.bufferWidth;
      }
      if (totalWidth < width + (scrollRate * 3)) {
        BufferArea previousBuffer = (BufferArea)buffers.get(buffers.size()-1);
        if (artikulationScoreCounter < 21 &&
          playScore) {
          artikulationImageScalers[artikulationScoreCounter] = PApplet.parseFloat(height) / PApplet.parseFloat(artikulationCropped[artikulationScoreCounter].height);
          println(artikulationImageScalers[artikulationScoreCounter]);
          buffers.add(new BufferArea(abs(PApplet.parseInt(artikulationCropped[artikulationScoreCounter].width * artikulationImageScalers[artikulationScoreCounter])), height, previousBuffer.scrollPosition + previousBuffer.bufferWidth));
          BufferArea buf = (BufferArea)buffers.get(buffers.size()-1);
          buf.displayImage = artikulationCropped[artikulationScoreCounter];
          artikulationScoreCounter++;
        } else if (playGenerative) {
          buffers.add(new BufferArea(PApplet.parseInt(random(min(500, width))) + minimumBufferWidth, height, previousBuffer.scrollPosition + previousBuffer.bufferWidth));
          buffersGenerated++;
        }
      }
    }
  }

  if (isPlaying) {
    //playback line
    if (buffersGenerated <=0) {
      strokeWeight(1);
      stroke(0, 30);
      line(width/2, 0, width/2, height);
    } else if (buffersGenerated > 0 &&
      buffersGenerated <= maxGeneratedBuffers) {
      strokeWeight(1);
      stroke(0, 30);
      line(width/2, 0, width/2, height);
    } else {
      strokeWeight(1);
      stroke(255, tiltTint);
      line(width/2, 0, width/2, height);
    }
  } else {
    //title text
    fill(0);
    textAlign(CENTER);
    textFont(titleFont, 32);
    text("Artikulation", width/2, height/2.5f);

    fill(0, 48);
    textFont(titleFont, 10);
    text("Gyorgy Ligeti 1958, Rainer Wehinger 1970, Kevin Schlei 2012", width/2, height - 20);


    fill(0);
    textAlign(CENTER);
    //    text("Artikulation", width/2, height/2.5);

    textFont(titleFont, 10);
    text("Spacebar: play all", width/2, height/2.0f + 60);
    text("S: play score only", width/2, height/2.0f + 75);
    text("G: play generative only", width/2, height/2.0f  + 90);

    PImage textureSmall = loadImage("texture-small.png");

    noStroke();
    float cx, ss, off;
    cx = width/2.0f;
    ss = 20;
    off = 11;
    for (int i=0; i<allColors.length; i++) {
      //      for (int j=0; j<2; j++) {
      fill(allColors[i]);
      float ox = (cx + (ss+off)*i) - ((ss+off)*allColors.length - off)/2.0f;
      float oy = height/2.0f;

      rect(ox, oy, ss, ss);

      tint(255, 75);
      image(textureSmall, ox, oy, ss, ss);
      tint(255);
    }
  }

  //  println(buffers.size());
  //  println(frameRate);
}

public void keyPressed() {
  //always clear buffer
  buffers.clear();
  artikulationScoreCounter = 0;

  println("keyCode "+keyCode);
  switch(keyCode) {
  case 83: //s
    {
      playScore = true;
      playGenerative = false;

      //add first buffer, make it score:
      artikulationImageScalers[artikulationScoreCounter] = PApplet.parseFloat(height) / PApplet.parseFloat(artikulationCropped[artikulationScoreCounter].height);
      println(artikulationImageScalers[artikulationScoreCounter]);
      buffers.add(new BufferArea(abs(PApplet.parseInt(artikulationCropped[artikulationScoreCounter].width * artikulationImageScalers[artikulationScoreCounter])), height, width));
      BufferArea buf = (BufferArea)buffers.get(buffers.size()-1);
      buf.displayImage = artikulationCropped[artikulationScoreCounter];
      artikulationScoreCounter++;

      isPlaying = true;
      loop();
    }
    break;
  case 71: //g
    {
      playGenerative = true;
      playScore = false;

      //add first buff generative
      buffers.add(new BufferArea(100, height, width));

      isPlaying = true;
      loop();
    }
    break;
  case 32: //space
    {
      playScore = true;
      playGenerative = true;

      //add first buffer, make it score:
      artikulationImageScalers[artikulationScoreCounter] = PApplet.parseFloat(height) / PApplet.parseFloat(artikulationCropped[artikulationScoreCounter].height);
      println(artikulationImageScalers[artikulationScoreCounter]);
      buffers.add(new BufferArea(abs(PApplet.parseInt(artikulationCropped[artikulationScoreCounter].width * artikulationImageScalers[artikulationScoreCounter])), height, width));
      BufferArea buf = (BufferArea)buffers.get(buffers.size()-1);
      buf.displayImage = artikulationCropped[artikulationScoreCounter];
      artikulationScoreCounter++;

      isPlaying = true;
      loop();
    }
    break;
  default:
    break;
  }

  isPlaying = true;
  loop();
}

class Block implements Shape {
  int dColor;
  int minXSize,minYSize,maxXSize,maxYSize;

  //Dot
  Block() {
    //defaults
    minXSize = 15;
    minYSize = 15;
    maxXSize = 25;
    maxYSize = 25;
  } //Dot

  //updateDraw
  public @Override
    void updateDraw(int dOriginX, int dOriginY, int dWidth, int dHeight, PGraphics buffer) {
    dColor = getDotColor();

    buffer.ellipseMode(LEFT);

    buffer.pushMatrix();
    buffer.translate(dOriginX, dOriginY);

    int dSize = min(dWidth, dHeight);

    if (showRects) {
      buffer.fill(120, 80, 0);
      buffer.stroke(0);
      buffer.rect(0, 0, dWidth, dHeight);
    }

    int horizPlay = dWidth - dSize;
    int vertPlay = dHeight - dSize;
    int targetX = PApplet.parseInt(random(horizPlay));
    int targetY = PApplet.parseInt(random(vertPlay));

    buffer.noStroke();
    buffer.fill(dColor);
    buffer.rect(targetX, targetY, dSize, dSize);

    buffer.popMatrix();
  } //updateDraw

  public void setSizes(int minX, int minY, int maxX, int maxY) {
    minXSize = minX;
    minYSize = minY;
    maxXSize = maxX;
    maxYSize = maxY;
  }
  
  public int minXSize() {
    return minXSize;
  }
  public int minYSize() {
    return minYSize;
  }
  public int maxXSize() {
    return maxXSize;
  }
  public int maxYSize() {
    return maxYSize;
  }
}
class Blocks implements Shape {
  int dColor, dColor2;
  int dTeeth;
  int dColumns;
  boolean dCheckered;
  int minXSize, minYSize, maxXSize, maxYSize;

  //Blocks
  Blocks() {
    //defaults:
    dTeeth = 5;
    //    dColumns = l;
  } //Blocks

  //updateDraw
  public @Override
    void updateDraw(int dOriginX, int dOriginY, int dWidth, int dHeight, PGraphics buffer) {

    dColor = getCombColor();
    dColor2 = getCombColor();

    buffer.pushMatrix();
    buffer.translate(dOriginX, dOriginY);

    if (showRects) {
      buffer.fill(120, 80, 0);
      buffer.noStroke();
      buffer.rect(0, 0, dWidth, dHeight);
    }

    buffer.fill(dColor);
    buffer.noStroke();

    float leftWidth = min(20, PApplet.parseFloat(dWidth) / 6.0f);
    leftWidth = max(leftWidth, 8);
    //    rect(0, 0, leftWidth, dHeight);

    //    dTeeth =  int(dHeight / 20);

    float toothYDivision = PApplet.parseFloat(dHeight) / PApplet.parseFloat(dTeeth);
    float toothHeight = toothYDivision * 0.76f;
    float makeupWidth = (toothYDivision * 0.14f) / (dTeeth - 1);
    float toothYPos = 0.0f;

    dColumns = PApplet.parseInt(dWidth / (toothHeight * 1.5f));
    dCheckered = random(1) > 0.66f;
    
    int teethLimited = PApplet.parseInt(random(dTeeth - 1)) + 2;
    
    //normal
    for (int i=0; i<teethLimited; i++) {
      for (int j=0; j<dColumns; j++) {
        toothYPos = toothYDivision * i + (makeupWidth * i);
        //        rect(0, toothYPos, dWidth, toothHeight);
        //        rect

        if (i%2==0 &&
          dCheckered) {
          buffer.fill(dColor2);
        } 
        else {
          buffer.fill(dColor);
        }
        buffer.noStroke();
        buffer.rect((toothHeight * 1.2f) * j + PApplet.parseInt(random(3)), toothYPos +  (j*0.05f * toothHeight), toothHeight * (1.0f - j*0.1f), toothHeight * (1.0f - j*0.1f));
      }
    }
    buffer.strokeWeight(1);


    buffer.popMatrix();
  } //updateDraw

  public void setSizes(int minX, int minY, int maxX, int maxY) {
    minXSize = minX;
    minYSize = minY;
    maxXSize = maxX;
    maxYSize = maxY;
  }
  public int minXSize() {
    return minXSize;
  }
  public int minYSize() {
    return minYSize;
  }
  public int maxXSize() {
    return maxXSize;
  }
  public int maxYSize() {
    return maxYSize;
  }
}
class BlocksTail implements Shape {
  int dColor, dColor2;
  int dTeeth;
  int dColumns;
  boolean dCheckered;
  int minXSize, minYSize, maxXSize, maxYSize;

  //Blocks
  BlocksTail() {
    //defaults:
    dTeeth = 5;
    //    dColumns = l;
  } //Blocks

  //updateDraw
  public @Override
    void updateDraw(int dOriginX, int dOriginY, int dWidth, int dHeight, PGraphics buffer) {

    dColor = getCombColor();
    dColor2 = getCombColor();

    buffer.pushMatrix();
    buffer.translate(dOriginX, dOriginY);

    if (showRects) {
      buffer.fill(120, 80, 0);
      buffer.noStroke();
      buffer.rect(0, 0, dWidth, dHeight);
    }

    buffer.fill(dColor);
    buffer.noStroke();

    float leftWidth = min(20, PApplet.parseFloat(dWidth) / 6.0f);
    leftWidth = max(leftWidth, 8);
    //    rect(0, 0, leftWidth, dHeight);

    //    dTeeth =  int(dHeight / 20);

    float toothYDivision = PApplet.parseFloat(dHeight) / PApplet.parseFloat(dTeeth);
    float toothHeight = toothYDivision * 0.76f;
    float makeupWidth = (toothYDivision * 0.14f) / (dTeeth - 1);
    float toothYPos = 0.0f;

    dColumns = PApplet.parseInt(dWidth / (toothHeight * 1.5f));
    dCheckered = random(1) > 0.66f;
    //normal
    for (int i=0; i<dTeeth; i++) {
        toothYPos = toothYDivision * i + (makeupWidth * i);

        if (i%2==0 &&
          dCheckered) {
          buffer.fill(dColor2);
        } 
        else {
          buffer.fill(dColor);
        }
        buffer.noStroke();
        float leftX = (toothHeight * 1.2f) + PApplet.parseInt(random(3));
        buffer.rect(leftX, toothYPos +  toothHeight, toothHeight, toothHeight);
        buffer.rect(leftX, toothYPos +  toothHeight + (toothHeight / 4.0f), dWidth - leftX, toothHeight - (toothHeight / 2.0f));

    }
    buffer.strokeWeight(1);


    buffer.popMatrix();
  } //updateDraw

  public void setSizes(int minX, int minY, int maxX, int maxY) {
    minXSize = minX;
    minYSize = minY;
    maxXSize = maxX;
    maxYSize = maxY;
  }
  public int minXSize() {
    return minXSize;
  }
  public int minYSize() {
    return minYSize;
  }
  public int maxXSize() {
    return maxXSize;
  }
  public int maxYSize() {
    return maxYSize;
  }
}
class BlocksTailGroup implements Shape {
  int minXSize, minYSize, maxXSize, maxYSize;

  //Blocks
  BlocksTailGroup() {
  } //Blocks

  //updateDraw
  public @Override
    void updateDraw(int dOriginX, int dOriginY, int dWidth, int dHeight, PGraphics buffer) {

    buffer.pushMatrix();
    buffer.translate(dOriginX, dOriginY);

      if (showRects) {
        buffer.fill(255, 255, 0);
        buffer.stroke(0);
        buffer.rect(0, 0, dWidth, dHeight);
      }

      //determine how many can really fit
      float area = dWidth * dHeight;
      int targetBlocksCount = PApplet.parseInt(area / 9000.0f);

      for (int i=0; i<targetBlocksCount; i++) {
        BlocksTail t = new BlocksTail();
        t.setSizes(50, 40, 150, 90);

        int ww = PApplet.parseInt(random(t.maxXSize() - t.minXSize())) + t.minXSize();
        int hh = PApplet.parseInt(random(t.maxYSize() - t.minYSize)) + t.minYSize();
        int yyo = PApplet.parseInt(random(dHeight - t.maxYSize()));
        int xxo = max(PApplet.parseInt(random(dWidth - t.maxXSize())),0);

        t.updateDraw(xxo, yyo, ww, hh, buffer);
      }

      buffer.strokeWeight(1);


      buffer.popMatrix();
  } //updateDraw

  public void setSizes(int minX, int minY, int maxX, int maxY) {
    minXSize = minX;
    minYSize = minY;
    maxXSize = maxX;
    maxYSize = maxY;
  }
  public int minXSize() {
    return minXSize;
  }
  public int minYSize() {
    return minYSize;
  }
  public int maxXSize() {
    return maxXSize;
  }
  public int maxYSize() {
    return maxYSize;
  }
}
//! normalized art size for comb, but not for dot swish?


class BufferArea {

  PGraphics buffer;
  PImage displayImage;
  float scrollPosition;
  int bufferWidth, bufferHeight;
  boolean displayScore;
  Shape sh;

  BufferArea(int w, int h, float s) {
    bufferWidth = w;
    bufferHeight = h;
    buffer = createGraphics(bufferWidth, bufferHeight, JAVA2D);
    scrollPosition = s;
    displayScore = false;

    update();
  }

  public void update() {

    //we just set displayImage when it's in score mode, no need to calculate
    if (!displayScore) {
      buffer.beginDraw();
      if (showRandBufferBackground) {
        buffer.background(PApplet.parseInt(random(255)), PApplet.parseInt(random(255)), PApplet.parseInt(random(255)));
      }      
      else {
        buffer.background(bgColor);
      }
      buffer.smooth();

      int objectsAdded = 0;

      int startingY = PApplet.parseInt(random(bufferHeight));
      startingY = max(startingY, 30);
      startingY = min(startingY, bufferHeight-30);

      if (random(1) < 0.5f &&
      bufferWidth > 160) {
        addReverbBlob(5, startingY);
      }

      objectsAdded = addShapes(0, startingY);

      if (objectsAdded == 0) {
        buffer.background(bgColor);
      }

      buffer.tint(255, 75);
      buffer.image(textureImage, 0, 0);
      buffer.tint(255);

      buffer.endDraw();

      displayImage = buffer.get();
    }
  }

  public void addReverbBlob(int startingX, int startingY) {
    if (bufferWidth > 120) {
      buffer.noStroke();
      buffer.fill(236, 227, 221);
      PVector ul, ur, ll, lr;
      float md = bufferWidth * 0.1f;
      ul = new PVector(random(50) + md, random((bufferHeight - md) / 2.0f) + md);
      ur = new PVector(bufferWidth - (random(50) + md), random((bufferHeight - md) / 2.0f) + md);
      ll = new PVector(random(50) + md, bufferHeight - (random((bufferHeight - md) / 2.0f) + md));
      lr = new PVector(bufferWidth - (random(50) + md), bufferHeight - (random((bufferHeight - md) / 2.0f) + md));

      if (ul.x > ur.x - 60) {
        ur.x = ul.x + 60;
      }
      if (ll.x > lr.x - 60) {
        lr.x = ll.x + 60;
      }
      if (ur.y > lr.y - 60) {
        lr.y = ur.y + 60;
      }
      if (ul.y > ll.y - 60) {
        ll.y = ul.y + 60;
      }

      buffer.beginShape();
      buffer.curveVertex(ur.x, ur.y);
      buffer.curveVertex(ul.x, ul.y);
      buffer.curveVertex(ll.x, ll.y);
      buffer.curveVertex(lr.x, lr.y);
      buffer.curveVertex(ur.x, ur.y);
      buffer.curveVertex(ul.x, ul.y);
      buffer.curveVertex(ll.x, ll.y);
      buffer.endShape();

      //buffer.noStroke();
      //buffer.fill(255,0,0);
      //buffer.ellipse(ul.x,ul.y,5,5);
      //buffer.ellipse(ur.x,ur.y,5,5);
      //buffer.ellipse(ll.x,ll.y,5,5);
      //buffer.ellipse(lr.x,lr.y,5,5);
    }
  }

  public int addShapes(int startingX, int startingY) {

    Shape sh, lowerShape, upperShape;

    sh = getMainShape();
    lowerShape = getLowerShape();
    upperShape = getUpperShape();

    int objectsAdded = 0;
    int xo, yo, w, h;
    xo = startingX;
    yo = startingY;


    w = PApplet.parseInt(random(sh.maxXSize() - sh.minXSize())) + sh.minXSize();
    h = PApplet.parseInt(random(sh.maxYSize() - sh.minYSize())) + sh.minYSize();

    //hacky:
//    println("maxX == "+sh.maxXSize());
    if (sh.maxXSize() > 400) { //catch blockGroups
//      println("it's a block group!");
      w = min(bufferWidth - xo, w);
    } 

    if (xo + w > bufferWidth) {
//      println("returning! too wide!");
      return 0;
    }
    while (yo + h > bufferHeight) {
      yo -= (yo + h) - bufferHeight + 1;
//      println("yo is now: "+yo+" h is: "+h);
    }
    buffer.stroke(255, 0, 0);
    buffer.fill(120, 54, 0);

    sh.updateDraw(xo, yo, w, h, buffer);
    objectsAdded++;

    while (true) {
      sh = getMainShape();

      //hacky:
      //      while (sh.maxXSize () > bufferWidth) { //catch BlocksTailGroup
      //        sh = getMainShape();
      //      }
      xo += w;
      yo += PApplet.parseInt(random(bufferHeight / 4)) - (bufferHeight / 8);
      w = PApplet.parseInt(random(sh.maxXSize() - sh.minXSize())) + sh.minXSize();
      h = PApplet.parseInt(random(sh.maxYSize() - sh.minYSize())) + sh.minYSize();

    if (sh.maxXSize() > 400) { //catch blockGroups
//      println("it's a block group inside the other loop!");
      w = min(bufferWidth - xo, w);
      
      if (w < 50) break;
    } 
    
      if (xo + w > bufferWidth) {
//        println("breaking! too wide!");
        break;
      }
      while (yo + h > bufferHeight) {
        yo -= h;
      }
      while (yo < 0) {
        yo += h;
      }

      sh.updateDraw(xo, yo, w, h, buffer);

      float coinFlip = random(1);
      if (coinFlip > 0.33f) {
        fill(255, 255, 0);
        int yyo = yo + h + PApplet.parseInt(random(lowerShape.minYSize()));
        int xxo = xo + PApplet.parseInt(random(w - lowerShape.minXSize()));
        int ww = PApplet.parseInt(random(lowerShape.maxXSize() - lowerShape.minXSize())) + lowerShape.minXSize();
        int hh = PApplet.parseInt(random(lowerShape.maxYSize() - lowerShape.minYSize())) + lowerShape.minYSize();
        if (yyo + hh < bufferHeight &&
          xxo + ww < bufferWidth) {
          lowerShape.updateDraw(xxo, yyo, ww, hh, buffer);
        }
      }
      coinFlip = random(1);
      if (coinFlip > 0.33f) {
        fill(255, 255, 0);

        int prevHH = 0;
        for (int i=0; i<2; i++) {
          int ww = PApplet.parseInt(random(upperShape.maxXSize() - upperShape.minXSize())) + upperShape.minXSize();
          int hh = PApplet.parseInt(random(upperShape.maxYSize() - upperShape.minYSize())) + upperShape.minYSize();
          int yyo = yo - (hh + prevHH) - PApplet.parseInt(random(upperShape.minYSize()));
          int xxo = xo + PApplet.parseInt(random(PApplet.parseFloat(w) - upperShape.minXSize()));

          if (yyo > 0 &&
            xxo + ww < bufferWidth) {
            upperShape.updateDraw(xxo, yyo, ww, hh, buffer);
          }

          prevHH = hh;
        }
      }
      objectsAdded++;
    }

    return objectsAdded;
  }
}
class CGRect {

  //all normalized values
  float originX,originY,sizeWidth,sizeHeight;

  CGRect (float _x, float _y, float _w, float _h) {
   originX = _x;
   originY = _y;
   sizeWidth = _w; 
   sizeHeight = _h;
  }
}
class Comb implements Shape {
  int dColor;
  int dTeeth;
  boolean dClosed;
  int minXSize, minYSize, maxXSize, maxYSize;

  //Comb
  Comb() {
    //defaults
    dTeeth = 5;
    dClosed = false;
  } //Comb

  //updateDraw
  public @Override
    void updateDraw(int dOriginX, int dOriginY, int dWidth, int dHeight, PGraphics buffer) {
    dColor = getCombColor();

    buffer.pushMatrix();
    buffer.translate(dOriginX, dOriginY);

    if (showRects) {
      buffer.fill(120, 80, 0);
      buffer.noStroke();
      buffer.rect(0, 0, dWidth, dHeight);
    }

    buffer.fill(dColor);
    buffer.noStroke();

    float leftWidth = min(20, PApplet.parseFloat(dWidth) / 6.0f);
    leftWidth = max(leftWidth, 8);
    buffer.rect(0, 0, leftWidth, dHeight);

    dTeeth = PApplet.parseInt(random(5)) + 5;

    float toothYDivision = PApplet.parseFloat(dHeight) / PApplet.parseFloat(dTeeth);
    float toothHeight = toothYDivision * 0.6f;
    float makeupWidth = (toothYDivision * 0.4f) / (dTeeth - 1);
    float toothYPos = 0.0f;

    //normal
    int slope = PApplet.parseInt(random(3));
    for (int i=0; i<dTeeth; i++) {
      toothYPos = toothYDivision * i + (makeupWidth * i);
      //        rect(0, toothYPos, dWidth, toothHeight);

      buffer.strokeWeight(toothHeight);
      buffer.noFill();
      buffer.stroke(dColor);

      if (dWidth > 25) {
        switch(slope) {
        case 0:
          {
            buffer.bezier(
            0 + (toothHeight / 2.0f), 
            toothYPos + (toothHeight / 2.0f) - 1, 
            dWidth - (toothHeight * 2), 
            toothYPos + (toothHeight / 2.0f), 
            dWidth - (toothHeight / 2.0f), 
            toothYPos + (toothHeight / 2.0f) + (dHeight / 3.0f), 
            dWidth - (toothHeight / 2.0f), 
            toothYPos + (toothHeight / 2.0f) + (dHeight / 3.0f));
          } 
          break;
        case 1:
          {
            buffer.bezier(
            0 + (toothHeight / 2.0f), 
            toothYPos + (toothHeight / 2.0f) - 1, 
            dWidth - (toothHeight * 2), 
            toothYPos + (toothHeight / 2.0f), 
            dWidth - (toothHeight / 2.0f), 
            toothYPos + (toothHeight / 2.0f) - (dHeight / 3.0f), 
            dWidth - (toothHeight / 2.0f), 
            toothYPos + (toothHeight / 2.0f) - (dHeight / 3.0f));
          }
          break;
        case 2:
          {
            buffer.line(0 + (toothHeight / 2.0f), toothYPos + (toothHeight / 2.0f), dWidth - (toothHeight / 2.0f), toothYPos + (toothHeight / 2.0f));
          }
          break;
        default:
          break;
        }
      }
      else {
        buffer.line(0 + (toothHeight / 2.0f), toothYPos + (toothHeight / 2.0f), dWidth - (toothHeight / 2.0f), toothYPos + (toothHeight / 2.0f));
      }

      //      buffer.line(
      //      0 + (toothHeight / 2.0), 
      //      toothYPos + (toothHeight / 2.0), 
      //      dWidth - (toothHeight / 2.0), 
      //      toothYPos + (toothHeight / 2.0));
      buffer.strokeWeight(1);
    }
    strokeWeight(1);

    if (dClosed) {
      buffer.noStroke();
      buffer.fill(dColor);
      buffer.rect(dWidth - leftWidth, 0, leftWidth + 1, dHeight);
    }

    buffer.popMatrix();
  } //updateDraw

  public void setSizes(int minX, int minY, int maxX, int maxY) {
    minXSize = minX;
    minYSize = minY;
    maxXSize = maxX;
    maxYSize = maxY;
  }
  public int minXSize() {
    return minXSize;
  }
  public int minYSize() {
    return minYSize;
  }
  public int maxXSize() {
    return maxXSize;
  }
  public int maxYSize() {
    return maxYSize;
  }
}
class Dot implements Shape {
  int dColor;
  boolean dIsHollow;
  int minXSize,minYSize,maxXSize,maxYSize;

  //Dot
  Dot() {
    //defaults
    dIsHollow = false;
    minXSize = 15;
    minYSize = 15;
    maxXSize = 25;
    maxYSize = 25;
  } //Dot

  //updateDraw
  public @Override
    void updateDraw(int dOriginX, int dOriginY, int dWidth, int dHeight, PGraphics buffer) {
    dColor = getDotColor();

    buffer.ellipseMode(LEFT);

    buffer.pushMatrix();
    buffer.translate(dOriginX, dOriginY);

    int dSize = min(dWidth, dHeight);

    if (showRects) {
      buffer.fill(120, 80, 0);
      buffer.stroke(0);
      buffer.rect(0, 0, dWidth, dHeight);
    }

    int horizPlay = dWidth - dSize;
    int vertPlay = dHeight - dSize;
    int targetX = PApplet.parseInt(random(horizPlay));
    int targetY = PApplet.parseInt(random(vertPlay));

    buffer.noStroke();
    buffer.fill(dColor);
    buffer.ellipse(targetX, targetY, dSize, dSize);


    if (dIsHollow) {
      buffer.fill(bgColor);
      float scaler = 0.55f;
      buffer.ellipse(targetX + ((1.0f - scaler) / 2.0f) * dSize, targetY + (((1.0f - scaler) / 2.0f) * dSize), dSize * scaler, dSize * scaler);
    } 

    buffer.popMatrix();
  } //updateDraw

  public void setSizes(int minX, int minY, int maxX, int maxY) {
    minXSize = minX;
    minYSize = minY;
    maxXSize = maxX;
    maxYSize = maxY;
  }
  
  public int minXSize() {
    return minXSize;
  }
  public int minYSize() {
    return minYSize;
  }
  public int maxXSize() {
    return maxXSize;
  }
  public int maxYSize() {
    return maxYSize;
  }
}
class DotCurvedTail implements Shape {
  int dColor;
  boolean dIsHollow;
  boolean dSlopeIsDownward;
  boolean dNoSlope;
  float dMaxEllipseSize; //0-1
  float dFatness; //0-1
  int minXSize, minYSize, maxXSize, maxYSize;

  //Dot
  DotCurvedTail() {
    //defaults
    dIsHollow = false;
    dSlopeIsDownward = true;
    dMaxEllipseSize = 0.4f;
    dFatness = 0.3f;
  } //Dot

  //updateDraw
  public @Override
    void updateDraw(int dOriginX, int dOriginY, int dWidth, int dHeight, PGraphics buffer) {
    dColor = getDotColor();

    //    dSlopeIsDownward = random(1) > 0.5;

    buffer.ellipseMode(LEFT);

    buffer.pushMatrix();
    buffer.translate(dOriginX, dOriginY);

    if (showRects) {
      buffer.fill(120, 80, 0);
      buffer.stroke(0);
      buffer.rect(0, 0, dWidth, dHeight);
    }

    float minEllipseSize = (dHeight * dMaxEllipseSize) / 4.0f;

    //tail
    int divi = max(dWidth, dHeight);
    int requiredEllipses = PApplet.parseInt((divi / minEllipseSize) * 6);

    float targetEndY;
    targetEndY = dHeight - minEllipseSize;

    dSlopeIsDownward = random(1) > 0.5f;
    dIsHollow = random(1) > 0.66f;

    for (int i=0; i<requiredEllipses; i++) {
      buffer.noStroke();
      buffer.fill(dColor);

      float divisor = dHeight / minEllipseSize;
      if (dFatness <= 0) {
        divisor = dHeight / minEllipseSize;
      } 
      else {
        divisor = ((pow(PApplet.parseFloat(i) / requiredEllipses, 1.0f/max(0.01f, 1.0f - dFatness))) * 3.0f + 1); //should be 1.0 - 4.0
      }
      float dSize = (dHeight * dMaxEllipseSize) / divisor;
      float dSizeHalf = dSize / 2.0f;
      float xPosition = (PApplet.parseFloat(i) / PApplet.parseFloat(requiredEllipses)) * (dWidth - dSize); 

      float yPosition = pow((PApplet.parseFloat(i) / PApplet.parseFloat(requiredEllipses)), 2) * (targetEndY);
      //      println(requiredEllipses+" ypos");

      if (dSlopeIsDownward) {
        buffer.ellipse(xPosition, yPosition, dSize, dSize);
      } 
      else {
        buffer.ellipse(xPosition, dHeight - yPosition - dSize, dSize, dSize);
      }
    }

    //main dot
    //    buffer.noStroke();
    //    buffer.fill(dColor);
    float dSize = dHeight * dMaxEllipseSize;
    //    if (dSlopeIsDownward) {
    //      buffer.ellipse(0, 0, dSize, dSize);
    //    } 
    //    else {
    //      buffer.ellipse(0, dHeight - dSize, dSize, dSize);
    //    }

    if (dIsHollow) {
      buffer.fill(bgColor);
      float scaler = 0.55f;
      if (dSlopeIsDownward) {
        buffer.ellipse(0 + (((1.0f - scaler) / 2.0f) * (dHeight * dMaxEllipseSize)), 0 + (((1.0f - scaler) / 2.0f) * (dHeight * dMaxEllipseSize)), dHeight * scaler * dMaxEllipseSize, dHeight * scaler * dMaxEllipseSize);
      } 
      else {
        buffer.ellipse(0 + (((1.0f - scaler) / 2.0f) * (dHeight * dMaxEllipseSize)), dHeight - (((1.0f - scaler) / 2.0f) * dSize) - (dSize * scaler), dSize * scaler, dSize * scaler);
      }
    } 
    buffer.popMatrix();
  } //updateDraw

  public void setSizes(int minX, int minY, int maxX, int maxY) {
    minXSize = minX;
    minYSize = minY;
    maxXSize = maxX;
    maxYSize = maxY;
  }
  public int minXSize() {
    return minXSize;
  }
  public int minYSize() {
    return minYSize;
  }
  public int maxXSize() {
    return maxXSize;
  }
  public int maxYSize() {
    return maxYSize;
  }
}
class DotStraightTail implements Shape {
  int dColor;
  boolean dIsHollow;
  float dFatness; //0-1
  int minXSize, minYSize, maxXSize, maxYSize;

  //Dot
  DotStraightTail() {
    dIsHollow = false;
    dFatness = 0.1f;
  } //Dot

  //updateDraw with buffer
  public @Override
    void updateDraw(int dOriginX, int dOriginY, int dWidth, int dHeight, PGraphics buffer) {
    dColor = getDotColor();

    buffer.ellipseMode(LEFT);

    buffer.pushMatrix();
    buffer.translate(dOriginX, dOriginY);

    if (showRects) {
      buffer.fill(120, 80, 0);
      buffer.stroke(0);
      buffer.rect(0, 0, dWidth, dHeight);
    }

    buffer.strokeWeight = 1.0f;

    //tail
    int requiredEllipses = PApplet.parseInt((dWidth / (dHeight / 4.0f)) * 4);
    for (int i=0; i<requiredEllipses; i++) {
      buffer.noStroke();
      buffer.fill(dColor);

      float divisor = 4.0f;
      if (dFatness <= 0) {
        divisor = 4.0f;
      } 
      else {
        divisor = ((pow(PApplet.parseFloat(i) / requiredEllipses, 1.0f/max(0.01f, 1.0f - dFatness))) * 3.0f + 1); //should be 1.0 - 4.0
      }
      float dSize = dHeight / divisor;
      float dSizeHalf = dSize / 2.0f;
      float xPosition = (PApplet.parseFloat(i) / PApplet.parseFloat(requiredEllipses)) * (dWidth - dSize); 

      buffer.ellipse(xPosition, (dHeight / 2.0f) - dSizeHalf, dSize, dSize);
    }


    //main dot
    buffer.noStroke();
    buffer.fill(dColor);
    buffer.ellipse(0, 0, dHeight, dHeight);

    if (dIsHollow) {
      buffer.fill(bgColor);
      float scaler = 0.55f;
      buffer.ellipse(0 + (((1.0f - scaler) / 2.0f) * dHeight), 0 + (((1.0f - scaler) / 2.0f) * dHeight), dHeight * scaler, dHeight * scaler);
    } 
    buffer.popMatrix();
  } //updateDraw with buffer

  public void setSizes(int minX, int minY, int maxX, int maxY) {
    minXSize = minX;
    minYSize = minY;
    maxXSize = maxX;
    maxYSize = maxY;
  }
  public int minXSize() {
    return minXSize;
  }
  public int minYSize() {
    return minYSize;
  }
  public int maxXSize() {
    return maxXSize;
  }
  public int maxYSize() {
    return maxYSize;
  }
}
class DotsConnected implements Shape {
  int dColor;
  boolean dIsHollow;
  int minXSize, minYSize, maxXSize, maxYSize;

  //!hollow not allowed!

  //Dot
  DotsConnected() {
    dIsHollow = false;
  } //Dot

  //updateDraw
  public @Override
    void updateDraw(int dOriginX, int dOriginY, int dWidth, int dHeight, PGraphics buffer) {
    dColor = getDotColor();

    buffer.ellipseMode(LEFT);

    buffer.pushMatrix();
    buffer.translate(dOriginX, dOriginY);

    if (showRects) {
      buffer.fill(120, 80, 0);
      buffer.stroke(0);
      buffer.rect(0, 0, dWidth, dHeight);
    }

    float ellipseTargetSize = random(8) + 12;
    int randSize = PApplet.parseInt(PApplet.parseFloat(dWidth) / ellipseTargetSize);
    float dSize = PApplet.parseFloat(dWidth) / PApplet.parseFloat(randSize);

    buffer.fill(dColor);
    buffer.noStroke();
    float startingY = random(dHeight - dSize);
    buffer.ellipse(0, startingY, dSize, dSize);

    if (dIsHollow) {
      buffer.fill(bgColor);
      float scaler = 0.55f;
      buffer.ellipse(((1.0f - scaler) / 2.0f) * dSize, (((1.0f - scaler) / 2.0f) * dSize) + startingY, dSize * scaler, dSize * scaler);
    } 

    float dSizeHalf = dSize / 2.0f;

    PVector p = new PVector(dSizeHalf, dSizeHalf + startingY);
    for (int i=1; i<randSize-1; i++) {
      //      float ellipseDrawScaler = random(0.4) + 0.6;
      //      float dDrawSize = dSize * ellipseDrawScaler;


      float targetY = random(dHeight - dSize);
      buffer.fill(dColor);
      buffer.noStroke();
      float randXOffset = random(dSize*0.7f);


      if (i * dSize + dSize + randXOffset < dWidth) {
        buffer.ellipse((i * dSize) + randXOffset, targetY, dSize, dSize);

        float dSizeOffset = 0.8f * dSize;

        PVector n = new PVector(i * dSize + dSizeHalf + randXOffset, targetY + dSizeHalf);
        float d = dist(p.x, p.y, n.x, n.y);
        float dx = (n.x-p.x);
        float dy = (n.y-p.y);

        buffer.pushMatrix();
        buffer.translate(p.x, p.y);
        buffer.noFill();
        buffer.stroke(dColor);
        buffer.strokeWeight(4);
        if (d > (2.2f * dSizeOffset)) { 
          buffer.line(dx * (dSizeOffset / d), dy * (dSizeOffset / d), n.x - p.x - (dx * (dSizeOffset / d)), n.y  - p.y - (dy * (dSizeOffset / d)));
        }
        buffer.popMatrix();
        p = n;

        buffer.strokeWeight(1);

        dIsHollow = random(1) > 0.66f;
        if (dIsHollow) {
          buffer.fill(bgColor);
          buffer.noStroke();
          float scaler = 0.55f;
          buffer.ellipse((i * dSize) + (((1.0f - scaler) / 2.0f) * dSize) + randXOffset, targetY + (((1.0f - scaler) / 2.0f) * dSize), dSize * scaler, dSize * scaler);
        }
      }
    }

    buffer.popMatrix();
  } //updateDraw

  public void setSizes(int minX, int minY, int maxX, int maxY) {
    minXSize = minX;
    minYSize = minY;
    maxXSize = maxX;
    maxYSize = maxY;
  }
  public int minXSize() {
    return minXSize;
  }
  public int minYSize() {
    return minYSize;
  }
  public int maxXSize() {
    return maxXSize;
  }
  public int maxYSize() {
    return maxYSize;
  }
}
public interface Shape {

  public void updateDraw(int dOriginX, int dOriginY, int dWidth, int dHeight, PGraphics buffer);
  public int minXSize();
  public int minYSize();
  public int maxXSize();
  public int maxYSize();
  public void setSizes(int minX, int minY, int maxX, int maxY);

}
class Wedges implements Shape {
  int dColor;
  int dTeeth;
  int minXSize, minYSize, maxXSize, maxYSize;

  //Wedges
  Wedges() {
    dTeeth = 5;
  } //Wedges

  //updateDraw
  public @Override
    void updateDraw(int dOriginX, int dOriginY, int dWidth, int dHeight, PGraphics buffer) {
    dColor = getDotColor();

    buffer.pushMatrix();
    buffer.translate(dOriginX, dOriginY);

    //containing rect:
    if (showRects) {
      buffer.fill(120, 80, 0);
      buffer.noStroke();
      buffer.rect(0, 0, dWidth, dHeight);
    }

    buffer.fill(dColor);
    buffer.noStroke();

    float toothYDivision = PApplet.parseFloat(dHeight) / PApplet.parseFloat(dTeeth);
    float toothHeight = toothYDivision * 0.6f;
    float makeupWidth = (toothYDivision * 0.4f) / (dTeeth - 1);
    float toothYPos = 0.0f;

      boolean reversed = random(1) > 0.75f;

    for (int i=0; i<dTeeth; i++) {
      toothYPos = toothYDivision * i;
      //        rect(0, toothYPos, dWidth, toothHeight);

      buffer.strokeWeight(toothHeight);
      buffer.noStroke();
      buffer.fill(dColor);
      float r = random(3);

      if (reversed) {
        buffer.triangle(dWidth - r, toothYPos, dWidth - dWidth, toothYPos + (toothYDivision / 2.0f), dWidth - r, toothYPos + toothYDivision - 2);
      } 
      else {
        buffer.triangle(0 + r, toothYPos, dWidth, toothYPos + (toothYDivision / 2.0f), 0 + r, toothYPos + toothYDivision - 2);
      }
    }
    buffer.strokeWeight(1);

    buffer.popMatrix();
  } //updateDraw

  public void setSizes(int minX, int minY, int maxX, int maxY) {
    minXSize = minX;
    minYSize = minY;
    maxXSize = maxX;
    maxYSize = maxY;
  }
  public int minXSize() {
    return minXSize;
  }
  public int minYSize() {
    return minYSize;
  }
  public int maxXSize() {
    return maxXSize;
  }
  public int maxYSize() {
    return maxYSize;
  }
}
  public void settings() {  size(1024, 320, P2D); }
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "--present", "--window-color=#666666", "--stop-color=#cccccc", "Artikulation_Player" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
